/*******************************************************************
 *	Author:	Peter van Sebille (peter@yipton.demon.co.uk)
 *	Based on code by Otfried Cheong
 *
 *******************************************************************/
#include <e32std.h>
#include <e32base.h>
#include <e32hal.h>
#include <e32rom.h>
#include <e32ver.h>
#include <e32svr.h>

#include "ekern.h"

class TCapsArLoV1
{
public:
	TVersion version;
};


// version information
const TInt KMajorVersionNumber=0;
const TInt KMinorVersionNumber=0;
const TInt KBuildVersionNumber=0;


class DArLoDriver : public DLogicalDevice
{
public:
	DArLoDriver();
	virtual TInt Install();
	virtual void GetCaps(TDes8 &aDes) const;
	virtual DLogicalChannel *CreateL();
};


class DArLo : public DLogicalChannel
{
public:
	DArLo(DLogicalDevice *aDevice);
protected:
	virtual void DoCancel(TInt aReqNo);
	virtual void DoRequest(TInt aReqNo,TAny *a1,TAny *a2);
	virtual TInt DoControl(TInt aFunction,TAny *a1,TAny *a2);
	virtual TInt OriginalDoControl(TInt aFunction, TAny* a1, TAny* a2);

};



DArLo::DArLo(DLogicalDevice *aDevice)
  : DLogicalChannel(aDevice)
{
//  SetBehaviour(0); // indicate we do not support asynchronous requests
}

void DArLo::DoCancel(TInt /* aReqNo */)
{
	INT3;
}

void DArLo::DoRequest(TInt /*aReqNo*/,TAny* /*a1*/,TAny* /*a2*/)
{
	INT3;
}


void FiddleDomainAccess()
{
#ifdef __MARM__
	asm("mov r0, #0xffffffff    ");
	asm("mcr p15, 0, r0, c3, c0 ");
#endif
}

#define __NAKED__ __declspec( naked )

__NAKED__ TInt DArLo::OriginalDoControl(TInt aFunction, TAny* a1, TAny* a2)
{
#ifdef __MARM__
  // handle any synchronous request function
  // can trash r0-r3 and ip
  // on entry:
  //   r0=this
  //   r1=aFunction
  //   r2=a1 (PhysMapAdr)
  //   r3=a2 (BootAdr)
  // on exit:
  //   r0=result
  //                                  

  asm("  cmp r1,#0");
  asm("  beq boot");

  // Examine an address
  // [R2] contains address to examine
  asm("  ldr r3,[r2]"); // get address to examine in R3
  asm("  ldr r1,[r3]"); // get value
  asm("  str r1,[r2]"); // store value back into [R2]
  asm("  mov  r0,#0");  // no error
  asm("  mov  pc,lr");  // return

  // Booting
  asm("boot: ");

  // turn off interrupts
  asm("  mrs r0, cpsr");
  asm("  bic r0, r0, #0xdf");
  asm("  orr r0, r0, #0xd3");  // disable FIQ, IRQ, set SVC mode
  asm("  msr cpsr,r0");

  // load r0 with physical map address
  asm("  mov r0, r2");

  // turn off MMU (write zero into MMU register)
  asm("  mov R1,#070");      // not quite zero---don't want 26 bit mode
  asm("  mcr p15,0,R1,c1,c1");

  // we have two instructions here before hell breaks lose!
  // jump to R3
  asm("  mov pc,R3");
  // will never get here
#endif
}


TInt DArLo::DoControl(TInt aFunction, TAny* a1, TAny* a2)
{
	INT3;
	if (aFunction == 0)
	{
		OriginalDoControl(aFunction, a1, a2);
	}
	else if (aFunction == 1)
	{
		TUint*		p = (TUint*) a1;

		FiddleDomainAccess();

		(*p) = *((TUint32*) *p);
	}
	return KErrNone;
}




DArLoDriver::DArLoDriver()
{
}

TInt DArLoDriver::Install()
{
	INT3;
	TPtrC name=_L("ARLO");       // our name, used by User::FreeLogicalDevice
	return SetName(&name);       // set our name and return error code
}

void DArLoDriver::GetCaps(TDes8 &aDes) const
{
	TVersion  version(KMajorVersionNumber, KMinorVersionNumber, KBuildVersionNumber);

	INT3;
	// standard capability-descriptor pattern
	TCapsArLoV1 b; // a structure
	b.version=version; // fill it with information
	aDes.FillZ(aDes.MaxLength()); // zero-pad the caller's descriptor
	aDes.Copy((TUint8 *)&b,Min(aDes.MaxLength(),sizeof(b)));
}

DLogicalChannel* DArLoDriver::CreateL()
{
	INT3;
	return new(ELeave) DArLo(this);
}


EXPORT_C DLogicalDevice *CreateLogicalDevice()
{
	INT3;
	DLogicalDevice *arlo = new(ELeave) DArLoDriver;
	return arlo;
}


// DLL entry point, required by all DLLs,
// called on thread or process attach/detach
GLDEF_C TInt E32Dll(TDllReason /*aReason*/)
{
  return KErrNone; // do nothing, return no error
}

